#include "SIPParser.h"

#undef CRLF // for remove warinig
#include <osipparser2/osip_parser.h>
#include <osipparser2/sdp_message.h>

// ----------------------------------------------------------------------------
// SDPConnect
// ----------------------------------------------------------------------------
class SDPConnect
{
public:
  IP ip;
  bool initialize(sdp_connection* connection)
  {
    ip = 0;
    if (connection == NULL) return false;
    if (connection->c_addr == NULL) return false;
    ip = str_to_ip(connection->c_addr);
    return true;
  }
};

// ----------------------------------------------------------------------------
// SDPConnectList
// ----------------------------------------------------------------------------
class SDPConnectList : public VDPList<SDPConnect*>
{
public:
  SDPConnect* find(SDPConnect* rhs)
  {
    for (int i = 0; i < count(); i++)
    {
      SDPConnect* res = at(i);
      if (res->ip == rhs->ip) return res;
    }
    return NULL;
  }
};

// ----------------------------------------------------------------------------
// SDPMediaAlt
// ----------------------------------------------------------------------------
class SDPMediaAlt
{
public:
  IP  ip;
  int port;
  bool initialize(sdp_attribute_t* attribute)
  {
    if (attribute == NULL) return false;
    string field = attribute->a_att_field;
    if (field != "alt") return false;
    {
      char* p = attribute->a_att_value;
      while (true)
      {
        string ip = p;
        size_t i = ip.find_first_of(" ");
        if(i == string::npos) return false;
        ip = ip.substr(0, i);

        if (VDNet::isIPV4Address(ip))
        {
          ip = str_to_ip(ip);
          while (*p != ' ' && *p != '\0') p++;
          if (*p == '\0') return false;
          port = atoi(p);
          return true;
        }
        p++;
        if (*p == '\0') break;
      }
    }
    return false;
  }
};

// ----------------------------------------------------------------------------
// SDPMediaAltList
// ----------------------------------------------------------------------------
class SDPMediaAltList : public VDPList<SDPMediaAlt*>
{
public:
  SDPMediaAlt* find(SDPMediaAlt* rhs)
  {
    for (int i = 0; i < count(); i++)
    {
      SDPMediaAlt* res = at(i);
      if (res->ip == rhs->ip && res->port == rhs->port) return res;
    }
    return NULL;
  }
};

// ----------------------------------------------------------------------------
// SDPMedia
// ----------------------------------------------------------------------------
class SDPMedia
{
public:
  int port;
  SDPMediaAltList mediaAltList;
  SDPConnectList  connectList;
public:
  bool initialize(sdp_media_t* media)
  {
    port = 0;
    if (media == NULL) return false;
    if (media->m_port != NULL)
    {
      port = atoi(media->m_port);
    }

    __node* attributeNode = media->a_attributes.node;
    while (attributeNode != NULL)
    {
      sdp_attribute_t* attribute = (sdp_attribute_t*)attributeNode->element;
      SDPMediaAlt* sdpMediaAlt = new SDPMediaAlt;
      if (sdpMediaAlt->initialize(attribute) && mediaAltList.find(sdpMediaAlt) == NULL)
      {
          mediaAltList.add(sdpMediaAlt);
      } else
        delete sdpMediaAlt;

      attributeNode = (__node*)attributeNode->next;
    }

    __node* connectionNode = media->c_connections.node;
    while (connectionNode != NULL)
    {
      sdp_connection* connection = (sdp_connection*)connectionNode->element;
      SDPConnect* sdpConnect = new SDPConnect;
      if (sdpConnect->initialize(connection) && connectList.find(sdpConnect) == NULL)
      {
        connectList.add(sdpConnect);
      } else
        delete sdpConnect;

      connectionNode = (__node*)connectionNode->next;
    }
    return true;
  }
};

// ----------------------------------------------------------------------------
// SDPMediaList
// ----------------------------------------------------------------------------
typedef VDPList<SDPMedia*> SDPMediaList;

// ----------------------------------------------------------------------------
// SIPConnectList
// ----------------------------------------------------------------------------
SIPConnectList::SIPConnectList()
{
}

SIPConnectList::~SIPConnectList()
{
}

bool SIPConnectList::exists(IP ip, int port)
{
  for (int i = 0; i < count(); i++)
  {
    SIPConnect* connect = at(i);
    if (connect->ip == ip && connect->port == port)
      return true;
  }
  return false;
}

SIPConnect* SIPConnectList::add(IP ip, int port)
{
  SIPConnect* connect = new SIPConnect;
  connect->ip = ip;
  connect->port = port;
  if (VDPList<SIPConnect*>::add(connect) == -1)
  {
    delete connect;
    return NULL;
  }
  return connect;
}

void SIPConnectList::addConnectList(SIPConnectList* src)
{
  for (int i = 0; i < src->count(); i++)
  {
    SIPConnect* connect = src->at(i);
    if (!exists(connect->ip, connect->port))
      add(connect->ip, connect->port);
  }
}

void SIPConnectList::logConnectList()
{
  string log = "";
  for (int i = 0; i < count(); i++)
  {
    SIPConnect* connect = at(i);
    log = log + ip_to_str(connect->ip) + "(" + int_to_str(connect->port) + ") ";
  }
  LOG_DEBUG("[SIPParser.cpp] SIPConnectList::logConnectList %s", log.c_str());
}

// ----------------------------------------------------------------------------
// SIPParser
// ----------------------------------------------------------------------------
SIPParser::SIPParser()
{
  from = "";
  to = "";
  cseq_method = "";
  connectList.clear();
}

SIPParser::~SIPParser()
{
  connectList.clear();
}

bool g_parser_init_ok = false;
bool SIPParser::parse(char* buf, int len)
{
  int res;

  osip_message_t* sip = NULL;
  sdp_message_t*  sdp = NULL;

  __node* bodyNode;
  osip_body_t* body;

  // --------------------------------------------------------------------------
  // Parse SIP(Header)
  // --------------------------------------------------------------------------
  if (!g_parser_init_ok)
  {
    res = parser_init();
    if (res != OSIP_SUCCESS)
    {
      LOG_ERROR("[SIPParser.cpp] SIPParser::parse parser_init return %d", res);
      goto _error;
    }
    g_parser_init_ok = true;
  }

  res = osip_message_init(&sip);
  if (res != OSIP_SUCCESS)
  {
    LOG_ERROR("[SIPParser.cpp] SIPParser::parse osip_message_init return %d", res);
    goto _error;
  }

  res = osip_message_parse(sip, buf, len);
  if (res != OSIP_SUCCESS)
  {
    LOG_ERROR("[SIPParser.cpp] SIPParser::parse osip_message_parse return %d", res);
    goto _error;
  }

  //
  // Parse to
  //
  if (sip->from != NULL)
  {
    if (sip->from->url != NULL)
    {
      if (sip->from->url->username != NULL)
        from = sip->from->url->username;
    }
    else if (sip->from->displayname != NULL)
    {
      from = sip->from->displayname;
    }
  }
  while (true)
  {
    size_t position = from.find("\"");
    if (position == string::npos) break;
    from.replace(position, 1, "");
  }

  //
  // Parse to
  //
  if (sip->to != NULL)
  {
    if (sip->to->url != NULL)
    {
      if (sip->to->url->username != NULL)
        to = sip->to->url->username;
    }
    else if (sip->to->displayname != NULL)
    {
      to = sip->to->displayname;
    }
  }
  while (true)
  {
    size_t position = to.find("\"");
    if (position == string::npos) break;
    to.replace(position, 1, "");
  }

  //
  // Parse cseq_method
  //
  if (sip->cseq != NULL)
  {
    if (sip->cseq->method != NULL)
      this->cseq_method = sip->cseq->method;
  }

  // --------------------------------------------------------------------------
  // Parse SDP(Body)
  // --------------------------------------------------------------------------
  res = sdp_message_init(&sdp);
  if (res != OSIP_SUCCESS)
  {
    LOG_ERROR("[SIPParser.cpp] SIPParser::parse sdp_message_init return %d", res);
  }

  bodyNode = sip->bodies.node;
  while (bodyNode != NULL)
  {
    body = (osip_body*)bodyNode->element;
    res = sdp_message_parse(sdp, body->body);
    if (res != OSIP_SUCCESS)
    {
      LOG_ERROR("[SIPParser.cpp] SIPParser::parse sdp_message_parse return %d", res);
      goto _error;
    }
    if (!parseSDP(sdp)) goto _error;
    bodyNode = (__node*)bodyNode->next;
  }

  if (sip != NULL) osip_message_free(sip);
  if (sdp != NULL) sdp_message_free(sdp);
  return true;  

_error:
  if (sip != NULL) osip_message_free(sip);
  if (sdp != NULL) sdp_message_free(sdp);
  return false;
}

bool SIPParser::parseSDP(void* _sdp)
{
  sdp_message_t* sdp = (sdp_message_t*) _sdp;

  SDPConnect   sdpConnect;
  SDPMediaList sdpMediaList;

  sdpConnect.initialize(sdp->c_connection);

  __node* mediaNode = sdp->m_medias.node;
  while (mediaNode != NULL)
  {
    sdp_media_t* media = (sdp_media_t*)mediaNode->element;
    SDPMedia* sdpMedia = new SDPMedia;
    sdpMedia->initialize(media);
    sdpMediaList.add(sdpMedia);

    mediaNode = (__node*)mediaNode->next;
  }

  //
  // Add to connectList using Connection Infornation(ip) and Media Description(port)
  //
  if (sdpConnect.ip != 0)
  {
    for (int i = 0; i < sdpMediaList.count(); i++)
    {
      SDPMedia* sdpMedia = sdpMediaList.at(i);
      if (sdpMedia->port != 0)
      {
        if (!connectList.exists(sdpConnect.ip, sdpMedia->port))
          connectList.add(sdpConnect.ip, sdpMedia->port);
      }
    }
  }

  //
  // Add to connectList using Media Description (alt)
  //
  for (int i = 0; i < sdpMediaList.count(); i++)
  {
    SDPMedia* sdpMedia = sdpMediaList.at(i);
    for (int j = 0; j < sdpMedia->mediaAltList.count(); j++)
    {
      SDPMediaAlt* mediaAlt = sdpMedia->mediaAltList.at(j);
      if (!connectList.exists(mediaAlt->ip, mediaAlt->port))
        connectList.add(mediaAlt->ip, mediaAlt->port);
    }

    if (sdpMedia->port != 0)
    {
      for (int j = 0; j < sdpMedia->connectList.count(); j++)
      {
        SDPConnect* sdpConnect = sdpMedia->connectList.at(j);
        if (!connectList.exists(sdpConnect->ip, sdpMedia->port))
          connectList.add(sdpConnect->ip, sdpMedia->port);
      }
    }
  }

  return true;
}
