#include "VSCall.h"

// ----------------------------------------------------------------------------
// VSCall
// ----------------------------------------------------------------------------
VSCall::VSCall()
{
  active = false;
  from = "unknown";
  to = "unknown";
  fromAddr.clear();
  toAddr.clear();
  callInfoVerified = false;

  memset(&begin_tv, 0, sizeof(begin_tv));
  memset(&end_tv, 0, sizeof(end_tv));
  memset(&begin_tm, 0, sizeof(begin_tm));
  memset(&end_tm, 0, sizeof(end_tm));

  packetCount        = 0;
  lastAccessTick     = 0;
  sessionList        = new VSSessionList;
  fromConnectList    = new SIPConnectList;
  toConnectList      = new SIPConnectList;
  unknownConnectList = new SIPConnectList;
}

VSCall::~VSCall()
{
  close();
  SAFE_DELETE(sessionList);
  SAFE_DELETE(fromConnectList);
  SAFE_DELETE(toConnectList);
  SAFE_DELETE(unknownConnectList);
}

void VSCall::setBeginTime(struct timeval* tv)
{
  this->begin_tv = *tv;
  this->begin_tm = *localtime(&tv->tv_sec);
}

void VSCall::setEndTime  (struct timeval * tv)
{
  this->end_tv = *tv;
  this->end_tm = *localtime(&tv->tv_sec);  
}

bool VSCall::open()
{
  if (active) return false;
  sessionList->clear();
  fromConnectList->clear();
  toConnectList->clear();
  unknownConnectList->clear();
  active = true;
  return true;
}

bool VSCall::close()
{
  if (!active) return false;
  active = false;
  setEndTime(&end_tv);
  sessionList->close();
  return true;
}

void VSCall::changeCallInfo(string from, string to, VSAddr* fromAddr, VSAddr* toAddr)
{
  this->from = from;
  this->to = to;
  this->fromAddr = *fromAddr;
  this->toAddr = *toAddr;
  callInfoVerified = true;
}

void VSCall::processPacket(SnoopIPPacket* packet)
{
  packetCount++;
  end_tv = packet->pktHdr->ts;
  lastAccessTick =
    packet->pktHdr->ts.tv_sec * 1000 +
    packet->pktHdr->ts.tv_usec / 1000;
}

// ----------------------------------------------------------------------------
// VSCallList
// ----------------------------------------------------------------------------
VSCallList::VSCallList()
{
}

VSCallList::~VSCallList()
{
  close();
}

bool VSCallList::open()
{
  clear();
  return true;
}

bool VSCallList::close()
{
  clear();
  return true;
}

VSCall* VSCallList::findCall(VSAddr* fromAddr, VSAddr* toAddr)
{
  for (int i = 0; i < count(); i++)
  {
    VSCall* call = at(i);
    if (call->fromAddr == *fromAddr && call->toAddr == *toAddr) return call;
  }
  return NULL;
}

VSSession* VSCallList::findSession(VSAddr* srcMAddr, VSAddr*dstMAddr)
{
  for (int i = 0; i < count(); i++)
  {
    VSCall* call = at(i);
    VSSession* session = call->sessionList->find(srcMAddr, dstMAddr);
    if (session != NULL) return session;
  }
  return NULL;
}

void VSCallList::getCallListByMediaAddr(VSAddr* mAddr, VSCallList* outCallList)
{
  for (int i = 0; i < count(); i++)
  {
    VSCall* call = at(i);
    if (call->fromConnectList->exists(mAddr->ip, mAddr->port))
      outCallList->add(call);
    if (call->toConnectList->exists(mAddr->ip, mAddr->port))
      outCallList->add(call);
    if (call->unknownConnectList->exists(mAddr->ip, mAddr->port))
      outCallList->add(call);
  }
}

void VSCallList::getDupCallList(VSCallList* callList1, VSCallList* callList2, VSCallList* outCallList)
{
  for (int i = 0; i < callList1->count(); i++)
  {
    VSCall* call1 = callList1->at(i);
    if (callList2->indexOf(call1) != NULL)
      outCallList->add(call1);
  }
}

VSCall* VSCallList::findCallByMediaAddr(VSAddr* mAddr1, VSAddr* mAddr2)
{
  VSCallList outCallList1, outCallList2, outCallListBoth;
  outCallList1.autoFree = false;
  outCallList2.autoFree = false;
  outCallListBoth.autoFree = false;

  getCallListByMediaAddr(mAddr1, &outCallList1);
  getCallListByMediaAddr(mAddr2, &outCallList2);
  getDupCallList(&outCallList1, &outCallList2, &outCallListBoth);

  if (outCallListBoth.count() > 0)
    return outCallListBoth.at(0);

  if (outCallList1.count() > 0)
    return outCallList1.at(0);

  if (outCallList2.count() > 0)
    return outCallList2.at(0);

  return NULL;
}

VSCall* VSCallList::findCallByConnectInfo(SIPConnect* connect)
{
  for (int i = 0; i < count(); i++)
  {
    VSCall* call = at(i);
    if (call->fromConnectList->exists(connect->ip, connect->port)) return call;
    if (call->toConnectList->exists(connect->ip, connect->port)) return call;
    if (call->unknownConnectList->exists(connect->ip, connect->port)) return call;
  }
  return NULL;
}

// ----------------------------------------------------------------------------
// VSCallMgr
// ----------------------------------------------------------------------------
void logSIPCommand(char* command, VSAddr* srcAddr, VSAddr* dstAddr)
{
  LOG_INFO("[VSCall.cpp] logSIPCommand %s detected %s > %s",
    command, srcAddr->str().c_str(), dstAddr->str().c_str());
}

bool VSCallMgr::process(
  VSSniff*       sniff,
  SnoopIPPacket* packet,
  VSAddr*        srcAddr,
  VSAddr*        dstAddr,              
  char*          data,
  int            dataLen)
{
  if (dataLen < 20) return false; // if len is too small

  VSCall* call;

  // --------------------------------------------------------------------------
  // INVITE (from -> to)
  // --------------------------------------------------------------------------
  if (strncmp(data, "INVITE", 6) == 0)
  {
    data[dataLen] = '\0';
    SIPParser parser;
    if (!parser.parse(data, dataLen)) return false;

    logSIPCommand("INVITE", srcAddr, dstAddr);

    call = sniff->callList->findCall(srcAddr, dstAddr);
    if (call == NULL)
    {
      LOG_INFO("[VSSession.cpp] VSSessionMgr::process create new call %s %s > %s %s",
        parser.from.c_str(), srcAddr->str().c_str(), parser.to.c_str(), dstAddr->str().c_str());
      call = sniff->createNewCall();
      call->setBeginTime(&packet->pktHdr->ts);
      call->changeCallInfo(parser.from, parser.to, srcAddr, dstAddr);
      if (!call->open())
      {
        LOG_ERROR("[VSCall.cpp] ******************************************");
        LOG_ERROR("[VSCall.cpp] VSCallMgr::process call->open return false");
        LOG_ERROR("[VSCall.cpp] ******************************************");
        delete call;
        return false;
      }
      sniff->callList->add(call);
    }
    if (parser.connectList.count() > 0)
    {
      call->fromConnectList->addConnectList(&parser.connectList);
      // call->fromConnectList->logConnectList(); // by gilgil 2009.02.22
    }

    return true;
  }

  // --------------------------------------------------------------------------
  // SIP/2.0 200 OK (from <- to)
  // --------------------------------------------------------------------------
  if (strncmp(data, "SIP/2.0 200", 11) == 0)
  {
    data[dataLen] = '\0';
    SIPParser parser;
    if (!parser.parse(data, dataLen)) return false;
    if (parser.cseq_method != "INVITE") return true; // not call connection

    logSIPCommand("SIP/2.0 200", srcAddr, dstAddr);

    call = sniff->callList->findCall(dstAddr, srcAddr);
    if (call == NULL)
    {
      for (int i = 0; i < parser.connectList.count(); i++)
      {
        SIPConnect* connect = parser.connectList.at(i);
        call = sniff->callList->findCallByConnectInfo(connect);
        if (call != NULL)
          break;
      }
    }
    if (call == NULL)
    {
      call = sniff->createNewCall();
      call->setBeginTime(&packet->pktHdr->ts);
      call->changeCallInfo(parser.from, parser.to, srcAddr, dstAddr);
      if (!call->open())
      {
        LOG_ERROR("[VSCall.cpp] ******************************************");
        LOG_ERROR("[VSCall.cpp] VSCallMgr::process call->open return false");
        LOG_ERROR("[VSCall.cpp] ******************************************");
        delete call;
        return false;
      }
      sniff->callList->add(call);
    }
    if (parser.connectList.count() > 0)
    {
      call->toConnectList->addConnectList(&parser.connectList);
      // call->toConnectList->logConnectList(); // by gilgil 2009.02.22
    }

    return true;
  }

  // --------------------------------------------------------------------------
  // CANCEL or BYE(from <-> to)
  // --------------------------------------------------------------------------
  char* closeCallMethod = NULL;
  if (strncmp(data, "CANCEL", 6) == 0) closeCallMethod = "CANCEL";
  if (strncmp(data, "BYE", 3) == 0)    closeCallMethod = "BYE";
  if (closeCallMethod != NULL)                              
  {
    data[dataLen] = '\0';
    SIPParser parser;
    if (!parser.parse(data, dataLen)) return false;

    logSIPCommand(closeCallMethod, srcAddr, dstAddr);

    call = sniff->callList->findCall(srcAddr, dstAddr);
    if (call != NULL)
      sniff->callList->remove(call);
    call = sniff->callList->findCall(dstAddr, srcAddr);
    if (call != NULL)
      sniff->callList->remove(call);
    return true;
  }

  return false;
}

