#include "VSSession.h"

// ----------------------------------------------------------------------------
// VSSession
// ----------------------------------------------------------------------------
VSSession::VSSession(VSCall* call)
{
  this->call = call;
  srcMAddr.clear();
  dstMAddr.clear();
  ssrc = 0;
  codec = NULL;
  bitBuffer = new VDBuffer;
  pcmBuffer = new VDBuffer;
}

VSSession::~VSSession()
{
  close();
  SAFE_DELETE(codec);
  SAFE_DELETE(bitBuffer);
  SAFE_DELETE(pcmBuffer);
}

#define BIG_ENOUGH_FRAME_COUNT 100
bool VSSession::open()
{
  if (codec == NULL) return false;
  if (!codec->open()) return false;
  bitBuffer->resize(codec->getDefaultBITSize() * BIG_ENOUGH_FRAME_COUNT);
  pcmBuffer->resize(codec->getDefaultPCMSize() * BIG_ENOUGH_FRAME_COUNT);
  return true;
}

bool VSSession::close()
{
  if (codec != NULL) codec->close();
  bitBuffer->clear();
  pcmBuffer->clear();
  return true;
}

void VSSession::processPayload(SnoopIPPacket* packet, char* payload, int payloadLen, bool* dataChanged)
{
  char* pcmMem  = pcmBuffer->mem();              
  int   pcmSize = pcmBuffer->size();
  int decodeLen = codec->decode(pcmMem, pcmSize, payload, payloadLen);
  if (decodeLen == -1)
  {
    LOG_ERROR("[VSSession.cpp] *************************************************");
    LOG_ERROR("[VSSession.cpp] VSSession::processPayload codec->decode return -1");
    LOG_ERROR("[VSSession.cpp] *************************************************");
    return;
  }
  processPCM(packet, pcmMem, decodeLen, payload, payloadLen, dataChanged);
  return;
}

// ----------------------------------------------------------------------------
// VSSessionList
// ----------------------------------------------------------------------------
VSSessionList::VSSessionList()
{
}

VSSessionList::~VSSessionList()
{
  close();
}

bool VSSessionList::open()
{
  clear();
  return true;
}

bool VSSessionList::close()
{
  clear();
  return true;
}

VSSession* VSSessionList::find(VSAddr* srcMAddr, VSAddr* dstMAddr)
{
  for (int i =0; i < count(); i++)
  {
    VSSession* session = at(i);
    if (session->srcMAddr == *srcMAddr && session->dstMAddr == *dstMAddr) return session;
  }
  return NULL;
}

// ----------------------------------------------------------------------------
// VSSessionMgr
// ----------------------------------------------------------------------------
VSSessionMgr::VSSessionMgr()
{
}

VSSessionMgr::~VSSessionMgr()
{
}

bool VSSessionMgr::process(
  VSSniff*       sniff,
  SnoopIPPacket* packet,
  VSAddr*        srcMAddr,
  VSAddr*        dstMAddr,
  char*          data,
  int            dataLen,
  bool*          dataChanged)
{
  if (dataLen < 20) return false; // if len is too small

	BYTE marker;
	BYTE payloadType;
	WORD seq;
	DWORD ts;
	DWORD ssrc;
	int payloadSize;
	char* payload;

  int res = VDRTPDeframeRTP(data, &marker, &payloadType, &seq, &ts, &ssrc, 0, &payload);
  if (res == -1) return false;
  payloadSize = dataLen - res;

  switch (payloadType)
  {
    case VDPT_G711:
    case VDPT_G723:
    case VDPT_G729:
      break;
    default:
      return false;
  }

  VSCall* call;
  VSSession* session;

  //
  // Find session
  //
  session = sniff->callList->findSession(srcMAddr, dstMAddr);
  if (session == NULL)
  {
    //
    // Find opposite session
    //
    VSSession* oppositeSession = sniff->callList->findSession(dstMAddr, srcMAddr);
    if (oppositeSession != NULL) // if other session exists
    {
      LOG_INFO("[VSSession.cpp] VSSessionMgr::process find call by opposite session %s > %s",
        srcMAddr->str().c_str(), dstMAddr->str().c_str());
      call = oppositeSession->call;
    } else
    {
      //
      // Find call by media addr
      //
      call = sniff->callList->findCallByMediaAddr(srcMAddr, dstMAddr);
      if (call == NULL)
      {
        //
        // Create new call
        //
        LOG_INFO("[VSSession.cpp] VSSessionMgr::process create unknown call %s > %s",
          srcMAddr->str().c_str(), dstMAddr->str().c_str());
        call = sniff->createNewCall();
        call->setBeginTime(&packet->pktHdr->ts);
        call->unknownConnectList->add(srcMAddr->ip, srcMAddr->port);
        call->unknownConnectList->add(dstMAddr->ip, dstMAddr->port);
        if (!call->open())
        {
          LOG_ERROR("[VSSession.cpp] *********************************************");
          LOG_ERROR("[VSSession.cpp] VSSessionMgr::process call->open return false");
          LOG_ERROR("[VSSession.cpp] *********************************************");
          delete call;
          return false;
        }
        sniff->callList->add(call);
      }
    }

    //
    // Create New Session
    //
    session = sniff->createNewSession(call);

    session->srcMAddr = *srcMAddr;
    session->dstMAddr = *dstMAddr;

    VSCodec* codec;
    switch (payloadType)
    {
      case VDPT_G711: codec = new VSCodecG711; break;
      case VDPT_G723: codec = new VSCodecG723; break;
      case VDPT_G729: codec = new VSCodecG729; break;
      default:
        LOG_ERROR("[VSSession.cpp] VSSessionMgr::process invalid PT(%d)", payloadType);
        delete session;
        return false;
    }
    session->codec = codec;

    session->ssrc = ssrc;

    if (!session->open())
    {
      LOG_ERROR("[VSSession.cpp] ************************************************");
      LOG_ERROR("[VSSession.cpp] VSSessionMgr::process session->open return false");
      LOG_ERROR("[VSSession.cpp] ************************************************");
      delete session;
      return false;
    }

    call->sessionList->add(session);
  }

  session->call->processPacket(packet);
  session->processPayload(packet, payload, payloadSize, dataChanged);
  return true;
}

