#include "VSSniffFile.h"
#include "vcl.h" // for DirectoryExists

// ----------------------------------------------------------------------------
// VSCallFile
// ----------------------------------------------------------------------------
VSCallFile::VSCallFile(VSSniffFile* sniffFile) : VSCall()
{
  this->sniffFile = sniffFile;
  callInfoChangedOnActive = false;
  waveFileName = "";
  wavRecord = new VDWAVRecord;
}

VSCallFile::~VSCallFile()
{
  close();
  SAFE_DELETE(wavRecord);
}

bool VSCallFile::open()
{
  if (!VSCall::open()) return false;

  if (sniffFile->saveWaveFile)
  {
    if (!wavRecord->active())
    {
      this->waveFileName = getWaveFileName();
      wavRecord->fileName = waveFileName;
      wavRecord->jitterTime = sniffFile->jitterTime;
      if (!wavRecord->open()) goto _error;
    }
  }

  return true;
_error:
  VSCall::close();
  return false;
}

bool VSCallFile::close()
{
  if (!VSCall::close()) return false;
  if (wavRecord->active()) wavRecord->close();

  if (packetCount < sniffFile->minCallPacketCount)
  {
    if (VDFile::fileExists(waveFileName))
      DeleteFile(waveFileName.c_str());
  }
  else
  if (callInfoChangedOnActive)
  {
    if (VDFile::fileExists(waveFileName))
    {
      string newWaveFileName = getWaveFileName();
      if (waveFileName != newWaveFileName)
      {
        BOOL res = MoveFile(waveFileName.c_str(), newWaveFileName.c_str());
        LOG_DEBUG("[VSSniffFile.cpp] VSCallFile::close filename MoveFile(%s, %s) return %d",
          waveFileName.c_str(), newWaveFileName.c_str(), res);
      }
    }
  }
  return true;
}

void VSCallFile::changeCallInfo(string from, string to, VSAddr* fromAddr, VSAddr* toAddr)
{
  VSCall::changeCallInfo(from, to, fromAddr, toAddr);
  if (active)
    callInfoChangedOnActive = true;
}

void VSCallFile::processPacket(SnoopIPPacket* packet)
{
  VSCall::processPacket(packet);
}

string VSCallFile::getWaveFileName()
{
  char dateTime[256];
  sprintf(dateTime, "%04d%02d%02d %02d%02d%02d",
    begin_tm.tm_year + 1900,
    begin_tm.tm_mon + 1,
    begin_tm.tm_mday,
    begin_tm.tm_hour,
    begin_tm.tm_min,
    begin_tm.tm_sec);

  string res;
  string pureFileName = sniffFile->waveFileFolder + + dateTime + " " + from + "_" + to;
  int i = 1;
  while (true)
  {
    res = pureFileName;
    if (i > 1) res = res + "_" + int_to_str(i);
    res = res + ".wav";
    if (!VDFile::fileExists(res)) break;
    i++;
  }
  return res;
}

// ----------------------------------------------------------------------------
// VSSessionFile
// ----------------------------------------------------------------------------
VSSessionFile::VSSessionFile(VSCallFile* callFile) : VSSession(callFile)
{
  this->callFile = callFile;
  callFile->wavRecord->add(this);
  mentBuffer = NULL;
  restMentBufferSize = 0;
}

VSSessionFile::~VSSessionFile()
{
  close();
}

bool VSSessionFile::open()
{
  bool res;

  if (!callFile->active)
  {
    res = callFile->open();
    if (!res) return false;
  }

  res = VSSession::open();
  if (callFile->sniffFile->mentBuffer != NULL)
  {
    this->mentBuffer = callFile->sniffFile->mentBuffer;
    this->restMentBufferSize = callFile->sniffFile->mentBufferSize;
  }
  return res;
}

bool VSSessionFile::close()
{
  return VSSession::close();
}

void VSSessionFile::processIPPacket(SnoopIPPacket* packet)
{
}

void VSSessionFile::processPCM(SnoopIPPacket* packet, char* pcm, int pcmLen, char* bit, int bitLen, bool* dataChanged)
{
  if (callFile->wavRecord->active())
  {
    DWORD tick =
      packet->pktHdr->ts.tv_sec * 1000 +
      packet->pktHdr->ts.tv_usec / 1000;
      callFile->wavRecord->write(this, tick, pcm, pcmLen);
  }

  if (restMentBufferSize > 0)
  {
    int size = pcmLen;
    if (size > restMentBufferSize) size = restMentBufferSize;
    int count = size / sizeof(short);
    short *src = (short*)mentBuffer;
    short *dst = (short*)pcm;

    int sum;
    for (int i = 0; i < count; i++)
    {
      sum = *src + *dst;
      if (sum < -32768) sum = -32768;
      else if (sum > 32767) sum = 32767;
      *dst = sum;
      src++;
      dst++;
    }
    mentBuffer += size;
    restMentBufferSize -= size;

    int encodeLen = codec->encode(bit, bitLen, pcm, pcmLen);
    if (encodeLen != -1) *dataChanged = true;
  }
}

// ----------------------------------------------------------------------------
// VSSniffFile
// ----------------------------------------------------------------------------
VSSniffFile::VSSniffFile(void* owner) : VSSniff(owner)
{
  string path = extractFilePath(application->exeName());

  saveCaptureFile = true;
  captureFileFolder = path + "cap\\";
  saveWaveFile = true;
  waveFileFolder = path + "wav\\";
  insertMent = true;
  mentFileName = "ment.wav";

  captureFileName = "";
  dump = new SnoopDump;
  mentBuffer = NULL;
  mentBufferSize = 0;

  jitterTime = 100; // 100 msec
}

VSSniffFile::~VSSniffFile()
{
  close();
  SAFE_DELETE(dump);
}

bool VSSniffFile::open()
{
  if (!VSSniff::open()) return false;

  bool res;

  //
  // captureFileFolder
  //
  if (saveCaptureFile)
  {
    if (!DirectoryExists(captureFileFolder.c_str()))
    {
      res = CreateDirectory(captureFileFolder.c_str(), NULL);
      if (!res)
      {
        LOG_ERROR("[VSSniffFile.cpp] VSSniff::open can not create folder(%s)", captureFileFolder.c_str());
        return false;
      }
    }

    captureFileName = getCaptureFileName();
    if (!dump->open(captureFileName)) return false;
  }

  //
  // waveFileFolder
  //
  if (saveWaveFile)
  {
    if (!DirectoryExists(waveFileFolder.c_str()))
    {
      res = CreateDirectory(waveFileFolder.c_str(), NULL);
      if (!res)
      {
        LOG_ERROR("[VSSniffFile.cpp] VSSniff::open can not create folder(%s)", waveFileFolder.c_str());
        return false;
      }
    }
  }

  //
  // mentBuffer
  //
  if (insertMent && mentFileName != "")
  {
    if (mentBuffer != NULL) delete[]mentBuffer;
    VDWaveFile waveFile;
    waveFile.file.mode = vdmRead;
    waveFile.file.fileName = mentFileName;
    if (!waveFile.open())
    {
      LOG_ERROR("[VSSniffFile.cpp] VSSniffFile::open waveFile.open(%s) return false", mentFileName.c_str());
      return false;
    }

    mentBufferSize = waveFile.file.getFileSize() - waveFile.headerSize;
    mentBuffer = new char[mentBufferSize];

    if (waveFile.read(mentBuffer, mentBufferSize) == VDERR_FAIL)
    {
      LOG_ERROR("[VDSniffFile.cpp] VDSniffFile::open waveFile.read return VDERR_FAIL");
      return false;
    }
    waveFile.close();
    LOG_DEBUG("[VSSniffFile.cpp] VSSniffFile::open mentBufferSize=%d", mentBufferSize);
  }

  return true;
}

bool VSSniffFile::close()
{
  bool res = VSSniff::close();
  if (dump->active()) dump->close();
  if (mentBuffer != NULL)
  {
    delete[] mentBuffer;
    mentBuffer = NULL;
  }
  mentBufferSize = 0;
  return res;
}

string VSSniffFile::getCaptureFileName()
{
  SYSTEMTIME st;
  GetLocalTime(&st);
  char dateTime[256];

  sprintf(dateTime, "%04d%02d%02d %02d%02d%02d",
    st.wYear, st.wMonth, st.wDay,  st.wHour, st.wMinute, st.wSecond);

  string res;
  string pureFileName = captureFileFolder + + dateTime;
  int i = 1;
  while (true)
  {
    res = pureFileName;
    if (i > 1) res = res + "_" + int_to_str(i);
    res = res + ".cap";
    if (!VDFile::fileExists(res)) break;
    i++;
  }

  return res;
}

void VSSniffFile::process(SnoopCapturer* capturer, SnoopIPPacket* packet)
{
  if (dump->active())
    dump->write(packet->pktHdr, (char*)packet->ethHdr);
  VSSniff::process(capturer, packet);
}

VSCall* VSSniffFile::createNewCall()
{
  VSCallFile* res = new VSCallFile(this);
  return res;
}

VSSession* VSSniffFile::createNewSession(VSCall* call)
{
  VSSession* res = new VSSessionFile((VSCallFile*)call);
  return res;
}

void VSSniffFile::load(VDXML* xml)
{
  VSSniff::load(xml);
  saveCaptureFile = xml->getBool("saveCaptureFile", saveCaptureFile);
  captureFileFolder = xml->getStr("captureFileFolder", captureFileFolder);
  saveWaveFile = xml->getBool("saveWaveFile", saveWaveFile);
  waveFileFolder = xml->getStr("waveFileFolder", waveFileFolder);
  insertMent = xml->getBool("insertMent", insertMent);
  mentFileName = xml->getStr("mentFileName", mentFileName);
  jitterTime = xml->getInt("jitterTime", jitterTime);
}

void VSSniffFile::save(VDXML* xml)
{
  VSSniff::save(xml);
  xml->setBool("saveCaptureFile", saveCaptureFile);
  xml->setStr("captureFileFolder", captureFileFolder);
  xml->setBool("saveWaveFile", saveWaveFile);
  xml->setStr("waveFileFolder", waveFileFolder);
  xml->setBool("insertMent", insertMent);
  xml->setStr("mentFileName", mentFileName);
  xml->setInt("jitterTime", jitterTime);  
}
